@@ -5,7 +5,7 @@ use strict;
 
 use vars qw($VERSION $AUTOLOAD $revision);
 
-$VERSION = '0.08';
+$VERSION = '0.13';
 $XML::XPathEngine::Namespaces = 0;
 $XML::XPathEngine::DEBUG = 0;
 
@@ -106,7 +106,11 @@ sub findnodes {
     if ($results->isa('XML::XPathEngine::NodeSet')) 
       { return wantarray ? $results->get_nodelist : $results; }
     else
-      { return wantarray ? () : XML::XPathEngine::NodeSet->new();   }
+      { return wantarray ? XML::XPathEngine::NodeSet->new($results) 
+                         : $results; 
+      } # result should be SCALAR
+      #{ return wantarray ? ($results) : $results; } # result should be SCALAR
+      #{ return wantarray ? () : XML::XPathEngine::NodeSet->new();   }
 }
 
 
@@ -116,14 +120,38 @@ sub findnodes_as_string {
     
     my $results = $self->find( $path, $context);
     
+
     if ($results->isa('XML::XPathEngine::NodeSet')) {
-        return join('', map { $_->toString } $results->get_nodelist);
+        return join '', map { $_->toString } $results->get_nodelist;
+    }
+    elsif ($results->isa('XML::XPathEngine::Boolean')) {
+        return ''; # to behave like XML::LibXML
     }
     elsif ($results->isa('XML::XPathEngine::Node')) {
         return $results->toString;
     }
     else {
-        return XML::XPathEngine::Node::XMLescape($results->value);
+        return _xml_escape_text($results->value);
+    }
+}
+
+sub findnodes_as_strings {
+    my $self = shift;
+    my ($path, $context) = @_;
+    
+    my $results = $self->find( $path, $context);
+    
+    if ($results->isa('XML::XPathEngine::NodeSet')) {
+        return map { $_->getValue } $results->get_nodelist;
+    }
+    elsif ($results->isa('XML::XPathEngine::Boolean')) {
+        return (); # to behave like XML::LibXML
+    }
+    elsif ($results->isa('XML::XPathEngine::Node')) {
+        return $results->getValue;
+    }
+    else {
+        return _xml_escape_text($results->value);
     }
 }
 
@@ -131,10 +159,22 @@ sub findvalue {
     my $self = shift;
     my ($path, $context) = @_;
     my $results = $self->find( $path, $context);
-    if ($results->isa('XML::XPathEngine::NodeSet')) { return $results->to_literal; }
-    return $results;
+    if ($results->isa('XML::XPathEngine::NodeSet')) 
+      { return $results->to_final_value; }
+      #{ return $results->to_literal; }
+    return $results->value;
 }
 
+sub findvalues {
+    my $self = shift;
+    my ($path, $context) = @_;
+    my $results = $self->find( $path, $context);
+    if ($results->isa('XML::XPathEngine::NodeSet')) 
+      { return $results->string_values; }
+    return ($results->string_value);
+}
+
+
 sub exists
 {
     my $self = shift;
@@ -160,23 +200,29 @@ sub set_var {
 sub set_namespace {
     my $self = shift;
     my ($prefix, $expanded) = @_;
+    $self->{uses_namespaces}=1;
     $self->{namespaces}{$prefix} = $expanded;
 }
 
 sub clear_namespaces {
     my $self = shift;
+    $self->{uses_namespaces}=0;
     $self->{namespaces} = {};
 }
 
 sub get_namespace {
     my $self = shift;
     my ($prefix, $node) = @_;
-    if (my $ns = $self->{namespaces}{$prefix}) {
-        return $ns;
-    }
-    if (my $nsnode = $node->getNamespace($prefix)) {
-        return $nsnode->getValue();
-    }
+   
+    my $ns= $node                    ? $node->getNamespace($prefix)
+          : $self->{uses_namespaces} ? $self->{namespaces}->{$prefix}
+          :                            $prefix;
+  return $ns;
+}
+
+sub set_strict_namespaces {
+    my( $self, $strict) = @_;
+    $self->{strict_namespaces}= $strict;
 }
 
 sub _get_context_set { $_[0]->{context_set}; }
@@ -189,9 +235,12 @@ sub _get_context_node { $_[0]->{context_set}->get_node($_[0]->{context_pos}); }
 sub _parse {
     my $self = shift;
     my $path = shift;
-    if ($CACHE{$path}) {
-        return $CACHE{$path};
-    }
+
+    my $context= join( '&&', $path, map { "$_=>$self->{namespaces}->{$_}" } sort keys %{$self->{namespaces}});
+    #warn "context: $context\n";
+
+    if ($CACHE{$context}) { return $CACHE{$context}; }
+
     my $tokens = $self->_tokenize($path);
 
     $self->{_tokpos} = 0;
@@ -201,8 +250,11 @@ sub _parse {
         # didn't manage to parse entire expression - throw an exception
         die "Parse of expression $path failed - junk after end of expression: $tokens->[$self->{_tokpos}]";
     }
-    
-    $CACHE{$path} = $tree;
+
+    $tree->{uses_namespaces}= $self->{uses_namespaces};   
+    $tree->{strict_namespaces}= $self->{strict_namespaces};   
+ 
+    $CACHE{$context} = $tree;
     
     _debug("PARSED Expr to:\n", $tree->as_string, "\n") if( $XML::XPathEngine::DEBUG);
     
@@ -914,6 +966,16 @@ sub _is_step {
       }
 }
 
+{ my %ENT;
+  BEGIN { %ENT= ( '&' => '&amp;', '<' => '&lt;', '>' => '&gt;', '"' => '&quote;'); }
+ 
+  sub _xml_escape_text
+    { my( $text)= @_;
+      $text=~ s{([&<>])}{$ENT{$1}}g;
+      return $text;
+    }
+}
+
 sub _debug {
     
     my ($pkg, $file, $line, $sub) = caller(1);
@@ -989,18 +1051,22 @@ In scalar context returns an XML::XPathEngine::NodeSet object.
 
 =head2 findnodes_as_string ($path, $context)
 
-Returns the nodes found reproduced as XML. The result is not guaranteed
-to be valid XML though.
+Returns the nodes found as a single string. The result is 
+not guaranteed to be valid XML though (it could for example be just text
+if the query returns attribute values).
+
+=head2 findnodes_as_strings ($path, $context)
+
+Returns the nodes found as a list of strings, one per node found.
 
 =head2 findvalue ($path, $context)
 
-Returns either a C<XML::XPathEngine::Literal>, a C<XML::XPathEngine::Boolean>
-or a C<XML::XPathEngine::Number> object. If the path returns a NodeSet,
-$nodeset->to_literal is called automatically for you (and thus a
-C<XML::XPathEngine::Literal> is returned). Note that
-for each of the objects stringification is overloaded, so you can just
-print the value found, or manipulate it in the ways you would a normal
-perl value (e.g. using regular expressions).
+Returns the result as a string (the concatenation of the values of the
+result nodes).
+
+=head2 findvalues($path, $context)
+
+Returns the values of the result nodes as a list of strings.
 
 =head2 exists ($path, $context)
 
@@ -1045,6 +1111,18 @@ Clears all previously set namespace mappings.
 
 Returns the uri associated to the prefix for the node (mostly for internal usage)
 
+=head2 set_strict_namespaces ($strict)
+
+By default, for historical as well as convenience reasons, XML::XPathEngine
+has a slightly non-standard way of dealing with the default namespace. 
+
+If you search for C<//tag> it will return elements C<tag>. As far as I understand it,
+if the document has a default namespace, this should not return anything.
+You would have to first do a C<set_namespace>, and then search using the namespace.
+
+Passing a true value to C<set_strict_namespaces> will activate this behaviour, passing a
+false value will return it to its default behaviour.
+
 =head2 set_var ($var. $val)
 
 Sets an XPath variable (that can be used in queries as C<$var>)
